# -*- coding: utf-8 -*-

from django.conf import settings
from django.core import checks, validators
from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.db.models import Q
from django.utils.encoding import smart_str
from PIL import Image
from shortuuid import uuid
from shortuuidfield import ShortUUIDField
import errno
import os


class AttachmentManager(models.Manager):
    """
    Manager for attachments (to upload a new attachment, mostly)
    """

    ATTACH_DIR = "attachments"
    ATTACH_PATH = os.path.join(settings.MEDIA_ROOT, ATTACH_DIR)
    THUMBNAIL_SIZE = (400, 400)
    IMAGE_EXTS = frozenset(["gif", "jpeg", "jpg", "png", "tif", "tiff"])
    USABLE_TYPES = frozenset(
        [
            "3gp",
            "7z",
            "aif",
            "aiff",
            "asf",
            "asx",
            "bmp",
            "cab",
            "cbr",
            "cda",
            "dmg",
            "doc",
            "eps",
            "flv",
            "gif",
            "gz",
            "hqx",
            "jpeg",
            "jpg",
            "m4a",
            "m4b",
            "m4v",
            "mid",
            "mov",
            "mp2",
            "mp3",
            "mp4",
            "mpeg",
            "mpg",
            "ogg",
            "pdf",
            "png",
            "pps",
            "ps",
            "rar",
            "rtf",
            "tgz",
            "tif",
            "tiff",
            "txt",
            "wav",
            "wma",
            "wmv",
            "wps",
            "xls",
            "zip",
        ]
    )
    MAX_UNUSED_ATTACHMENTS = 10

    def upload_file(self, namespace, uploaded_file):
        """
        Upload a file to the specified namespace.
        """

        file_name = uploaded_file.name
        base_name, extension = os.path.splitext(file_name)
        extension = extension[1:].lower()

        if extension not in self.USABLE_TYPES:
            raise ValidationError(f"L’extension « {str(extension)} » n’est pas autorisée.")

        if len(file_name) > 64:
            file_name = f"{base_name[:63 - len(extension)]}.{extension}"

        attachment = Attachment()
        attachment.file_name = file_name
        attachment.file_type = extension
        attachment.namespace = smart_str(namespace)

        attachment.save_with_file(uploaded_file)

        self._garbage_collect()

        return attachment

    def get_available_attachments_for(self, namespace, owner_id=0):
        """
        Returns the available attachments for the specified namespace and
        the specified owner ID: Attachments currently associated to this owner,
        and unassociated attachments.
        """

        # Return the newest files first, starting with all non-associated
        # files (for convenience)

        return self.filter(Q(namespace=namespace) & (Q(owner_id=0) | Q(owner_id=owner_id))).order_by(
            "owner_id", "-last_update"
        )

    def update_attachments_for(self, namespace, owner_id, attachment_idents, commit):
        """
        Sets the owner_id for the specified attachments IDs.
        This method checks that the attachment IDs exist in the specified
        namespace and that they are not associated with another owner.
        ValidationErrors are raised if it’s not the case.
        If commit is True, the changes are saved to the database; if it’s not,
        the checks are performed but changes are not saved.
        """

        attachment_idents = set(attachment_idents)

        if commit:
            with transaction.atomic():
                to_reset = self.filter(namespace=namespace, owner_id=owner_id)
                to_reset.update(owner_id=0)
                to_set = self.filter(namespace=namespace, owner_id=0, ident_main__in=attachment_idents)
                updated = to_set.update(owner_id=owner_id)

                if updated != len(attachment_idents):
                    can_be_set = self.filter(namespace=namespace, ident_main__in=attachment_idents)
                    can_be_set = set(can_be_set.values_list("ident_main", flat=True))
                    in_error = attachment_idents - can_be_set
                    assert in_error

                    if len(in_error) == 1:
                        raise ValidationError(f"La pièce jointe {str(in_error.pop())} n’est pas disponible.")
                    else:
                        raise ValidationError(
                            "Les pièces jointes {} ne sont pas disponibles.".format(",".join(in_error))
                        )
        else:
            to_set = self.filter(
                Q(namespace=namespace, ident_main__in=attachment_idents) & (Q(owner_id=0) | Q(owner_id=owner_id))
            )
            can_be_set = set(to_set.values_list("ident_main", flat=True))
            in_error = attachment_idents - can_be_set

            if len(in_error) == 1:
                raise ValidationError(f"""La pièce jointe {str(in_error.pop())} n’est pas disponible.""")
            elif len(in_error) > 1:
                raise ValidationError("Les pièces jointes {} ne sont pas disponibles.".format(",".join(in_error)))

    def _garbage_collect(self):
        """
        Remove oldest unused attachments.
        """

        query = self.filter(owner_id=0).order_by("-last_update")

        for attachment in query[self.MAX_UNUSED_ATTACHMENTS :]:
            attachment.delete()

    def check(self, **kwargs):
        """
        Django system checks implementation
        """

        errors = super(AttachmentManager, self).check(**kwargs)

        if os.path.normpath(settings.MEDIA_ROOT) == ".":
            errors.append(
                checks.Error(
                    "MEDIA_ROOT is not set (or set to .)",
                    hint=None,
                    obj=self,
                    id="attachments.E001",
                )
            )

        elif not os.path.isdir(self.ATTACH_PATH):
            errors.append(
                checks.Error(
                    "Attachment storage folder not found.",
                    hint=f"Create the storage folder {self.ATTACH_PATH}",
                    obj=self,
                    id="attachments.E002",
                )
            )

        else:
            file_path = os.path.join(self.ATTACH_PATH, "_test")
            try:
                fd = os.open(file_path, os.O_WRONLY | os.O_CREAT | os.O_EXCL)
            except OSError:
                errors.append(
                    checks.Error(
                        "Attachment storage folder not writable.",
                        hint=f"Make sure {self.ATTACH_PATH} is writable",
                        obj=self,
                        id="attachments.E003",
                    )
                )
            else:
                os.close(fd)
                os.unlink(file_path)

        return errors


class Attachment(models.Model):
    """
    Represents a file attachment.
    """

    objects = AttachmentManager()

    # Human-readable file name (original file name)
    file_name = models.CharField(max_length=64)

    # Identifier for the original file
    ident_main = ShortUUIDField(auto=False, primary_key=True)

    # Identifier for the miniature image file (for images)
    ident_mini = ShortUUIDField(unique=True, auto=False, null=True, blank=True)

    # File type (extension)
    file_type = models.CharField(max_length=4, null=False)

    # File size (in bytes)
    file_size = models.PositiveIntegerField(blank=True, null=False)

    # File namespace (indicate which categories of objects can manipulate
    # this file)
    namespace = models.CharField(
        max_length=32,
        null=False,
        blank=False,
        validators=[validators.RegexValidator(r"^[a-z0-9_]+$")],
    )

    # Owner ID (ID of the object currently owning the file — 0 if none)
    owner_id = models.IntegerField(null=False, default=0)

    # Last update date (used for garbage collection)
    last_update = models.DateTimeField(auto_now=True)

    def save_with_file(self, uploaded_file):
        """
        Saves the attachment to the database, with the uploaded_file as file
        data.
        """

        self.ident_main = uuid()

        is_image = self.file_type in AttachmentManager.IMAGE_EXTS

        # Parameter validation (so we don’t create a directory if a bogus
        # namespace is provided)
        self.full_clean()

        # Create the namespace storage directory
        try:
            os.mkdir(self._get_directory_path())
        except OSError as err:
            if err.errno != errno.EEXIST:
                raise

        saved = False
        try:
            with open(self.get_main_path(), "wb+") as main_fd:
                for chunk in uploaded_file.chunks():
                    main_fd.write(chunk)

                self.file_size = main_fd.tell()
                main_fd.seek(0)

                if is_image:
                    try:
                        image = Image.open(main_fd)
                        image.load()
                    except EnvironmentError:
                        raise ValidationError("Le fichier envoyé n’est pas " "une image valide.")

                    image.thumbnail(AttachmentManager.THUMBNAIL_SIZE, Image.ANTIALIAS)
                    self.ident_mini = uuid()
                    mini_path = self.get_mini_path()
                    with open(mini_path, "wb") as mini_fd:
                        image.save(mini_fd)
            self.save()
            saved = True
        finally:
            if not saved:
                # Exception occurred during the process — get rid of the files
                self._delete_files()

    def get_absolute_url(self):
        """
        Absolute URL to the main attachment
        """

        format = "{media_url}{attach_dir}/{namespace}/{ident_main}.{file_type}"
        return format.format(
            media_url=settings.MEDIA_URL,
            attach_dir=AttachmentManager.ATTACH_DIR,
            namespace=self.namespace,
            ident_main=self.ident_main,
            file_type=self.file_type,
        )

    def get_mini_url(self):
        """
        Absolute URL to the miniature, if any
        """

        format = "{media_url}{attach_dir}/{namespace}/{ident_mini}.{file_type}"
        return format.format(
            media_url=settings.MEDIA_URL,
            attach_dir=AttachmentManager.ATTACH_DIR,
            namespace=self.namespace,
            ident_mini=self.ident_mini,
            file_type=self.file_type,
        )

    def get_main_path(self):
        """
        Full path to the main attachment
        """

        return os.path.join(
            AttachmentManager.ATTACH_PATH,
            self.namespace,
            f"{self.ident_main}.{self.file_type}",
        )

    def get_mini_path(self):
        """
        Full path to the miniature, if any
        """

        assert self.ident_mini is not None

        return os.path.join(
            AttachmentManager.ATTACH_PATH,
            self.namespace,
            f"{self.ident_mini}.{self.file_type}",
        )

    def _get_directory_path(self):
        """
        Path to the namespace storage directory
        """

        return os.path.join(AttachmentManager.ATTACH_PATH, self.namespace)

    def delete(self, *args, **kwargs):
        """
        Database deletion function. Also deletes the attached files.
        """

        self._delete_files()
        super(Attachment, self).delete(*args, **kwargs)

    def _delete_files(self):
        """
        Erases the attached files.
        """

        if self.ident_main is not None:
            self._delete_if_exists(self.get_main_path())

        if self.ident_mini:
            self._delete_if_exists(self.get_mini_path())

        # Delete the namespace storage directory if empty
        try:
            os.rmdir(self._get_directory_path())
        except OSError as err:
            if err.errno not in {errno.ENOENT, errno.ENOTEMPTY}:
                raise

    @staticmethod
    def _delete_if_exists(path):
        """
        Deletes the file at the specified path if it exists.
        """

        try:
            os.unlink(path)
        except OSError as err:
            if err.errno != errno.ENOENT:
                raise

    def __str__(self):
        return f"{self.ident_main}.{self.file_type} owner {self.owner_id}"
